@@ -3,7 +3,7 @@
 """This module provides a simplified interface for running Glimpse experiments.
 
 The easiest way to use this module is via the top-level functions, such as
-:func:`SetCorpus`, which provide a declarative interface similar to
+:func:`SetCorpus`, which provide a declarative interface in the style of
 Matlab(TM). Alternatively, an object-oriented interface is also available by
 using the :class:`Experiment` class directly.
 
@@ -15,39 +15,48 @@ using the :class:`Experiment` class directly.
 # Please see the file COPYING in this distribution for usage
 # terms.
 
+import logging
+import math
+import numpy as np
+import operator
+import os
+import sklearn.metrics
+import sklearn.pipeline
+import sklearn.preprocessing
+import sklearn.svm
+import sys
+import time
+import types
+
 from glimpse import backends
 from glimpse.backends import InsufficientSizeException
 from glimpse.models import viz2
+from glimpse.models import misc
 from glimpse import pools
 from glimpse import util
 from glimpse.util import docstring
 from glimpse.util.grandom import HistogramSampler
 from glimpse.util import svm
 from glimpse.models.misc import InputSourceLoadException
-import logging
-import numpy as np
-import os
-import sys
-import time
+import glimpse.models
 
-__all__ = ( 'SetPool', 'UseCluster', 'SetModelClass', 'SetParams', 'GetParams',
-    'MakeModel', 'GetExperiment', 'SetExperiment', 'ImprintS2Prototypes',
-    'MakeUniformRandomS2Prototypes', 'MakeShuffledRandomS2Prototypes',
-    'MakeHistogramRandomS2Prototypes', 'MakeNormalRandomS2Prototypes',
-    'SetS2Prototypes', 'SetCorpus', 'SetTrainTestSplit', 'SetLayer',
-    'SetTrainTestSplitFromDirs', 'ComputeFeatures', 'CrossValidateSvm',
-    'TrainSvm', 'TestSvm', 'RunSvm', 'LoadExperiment', 'StoreExperiment',
-    'Reset', 'Verbose')
+__all__ = (
+    'ComputeFeatures', 'CrossValidateSvm', 'GetExampleCorpus',
+    'GetExampleImage', 'GetExampleImages', 'GetExperiment',
+    'GetImageFeatures', 'GetParams', 'ImprintS2Prototypes', 'LoadExperiment',
+    'MakeHistogramRandomS2Prototypes', 'MakeModel',
+    'MakeNormalRandomS2Prototypes', 'MakeShuffledRandomS2Prototypes',
+    'MakeUniformRandomS2Prototypes', 'SetCorpus', 'SetExperiment', 'SetLayer',
+    'SetModelClass', 'SetParams', 'SetPool', 'SetS2Prototypes',
+    'SetTrainTestSplit', 'SetTrainTestSplitFromDirs', 'StoreExperiment',
+    'Reset', 'RunSvm', 'TestSvm', 'TrainSvm', 'UseCluster', 'Verbose',
+    )
 
 class DirReader(object):
   """Read directory contents."""
 
   def __init__(self, ignore_hidden = True):
-    if ignore_hidden:
-      path_filter = DirReader._HiddenPathFilter
-    else:
-      path_filter = None
-    self.path_filter = path_filter
+    self.ignore_hidden = ignore_hidden
 
   @staticmethod
   def _HiddenPathFilter(path):
@@ -55,8 +64,10 @@ class DirReader(object):
     return not path.startswith('.')
 
   def _Read(self, dir_path):
-    return [ os.path.join(dir_path, entry)
-        for entry in filter(self.path_filter, os.listdir(dir_path)) ]
+    entries = os.listdir(dir_path)
+    if self.ignore_hidden:
+      entries = filter(DirReader._HiddenPathFilter, entries)
+    return [ os.path.join(dir_path, entry) for entry in entries ]
 
   def ReadDirs(self, dir_path):
     """Read list of sub-directories."""
@@ -74,7 +85,7 @@ class Experiment(object):
 
   """
 
-  def __init__(self, model, layer, pool, scaler):
+  def __init__(self, model, layer, pool):
     """Create a new experiment.
 
     :type model: model object
@@ -82,22 +93,18 @@ class Experiment(object):
     :param LayerSpec layer: The layer activity to use for features vectors.
     :type pool: pool object
     :param pool: A serializable worker pool.
-    :param svm.Scaler scaler: Feature scaling algorithm.
 
     """
     # Default arguments should be chosen in SetExperiment()
     assert model != None
     assert layer != None
     assert pool != None
-    assert scaler != None
     #: The Glimpse model used to compute feature vectors.
     self.model = model
     #: The worker pool used for parallelization.
     self.pool = pool
     #: The model layer from which feature vectors are extracted.
     self.layer = layer
-    #: The algorithm used to scale feature vectors.
-    self.scaler = scaler
     # Initialize attributes used by an experiment
     #: (list of str) Names of image classes.
     self.classes = []
@@ -125,11 +132,9 @@ class Experiment(object):
     #: (list of 2D ndarray) Feature vectors for the test images, indexed by
     #: class, image, and then feature offset.
     self.test_features = None
-    #: (dict) SVM evaluation results on the training data, in the format
-    #: returned by :func:`Svm.Test <glimpse.util.svm.Svm.Test>`.
+    #: (dict) SVM evaluation results on the training data.
     self.train_results = None
-    #: (dict) SVM evaluation results on the test data, in the format returned by
-    #: :func:`Svm.Test <glimpse.util.svm.Svm.Test>`.
+    #: (dict) SVM evaluation results on the test data.
     self.test_results = None
     #: (bool) Flag indicating whether cross-validation was used to compute test
     #: accuracy.
@@ -140,12 +145,38 @@ class Experiment(object):
     self.svm_train_time = None
     #: (float) Time required to test the SVM, in seconds.
     self.svm_test_time = None
-    self.debug = False
+    #: The file-system reader.
     self.dir_reader = DirReader(ignore_hidden = True)
-    self.debug_prototype_locations = None
+    #: (2D list of 4-tuple) Patch locations for imprinted prototypes.
+    self.prototype_imprint_locations = None
+    #: (float) Time elapsed while training and testing the SVM (in seconds).
     self.svm_time = None
+    #: (float) Time elapsed while constructing feature vectors (in seconds).
     self.compute_feature_time = None
 
+  def __eq__(self, other):
+    if other == None:
+      return False
+    attrs = dir(self)
+    attrs = [ a for a in attrs if not (a.startswith('_') or
+        isinstance(getattr(self, a), types.MethodType) or
+        isinstance(getattr(self, a), types.FunctionType)) ]
+    attrs = set(attrs) - set(('model', 'pool', 'dir_reader', 'classifier'))
+    for a in attrs:
+      value = getattr(self, a)
+      other_value = getattr(other, a)
+      if a in ('test_features', 'train_features', 's2_prototypes'):
+        test = util.CompareArrayLists(other_value, value)
+      elif a in ('test_results', 'train_results'):
+        test = all(other_value['predicted_labels'] == value['predicted_labels']) \
+            and all(other_value['decision_values'] == value['decision_values']) \
+            and other_value['accuracy'] == value['accuracy']
+      else:
+        test = other_value == value
+      if not test:
+        return False
+    return True
+
   def GetFeatures(self):
     """The full set of features for each class, without training/testing splits.
 
@@ -220,14 +251,16 @@ class Experiment(object):
     if self.train_images == None:
       sys.exit("Please specify the training corpus before imprinting "
           "prototypes.")
+    model = self.model
     start_time = time.time()
     image_files = self.train_images[0]
     # Represent each image file as an empty model state.
-    input_states = [ self.model.MakeStateFromFilename(f, resize = self.resize)
-        for f in image_files ]
+    input_states = [ model.MakeStateFromFilename(fn, resize = self.resize)
+        for fn in image_files ]
     try:
-      prototypes, locations = self.model.ImprintS2Prototypes(num_prototypes,
-          input_states, normalize = True, pool = self.pool)
+      prototypes, locations = misc.ImprintKernels(model, model.LayerClass.C1,
+          model.s2_kernel_sizes, num_prototypes, input_states,
+          normalize = model.s2_kernels_are_normed, pool = self.pool)
     except InputSourceLoadException, ex:
       if ex.source != None:
         path = ex.source.image_path
@@ -244,25 +277,35 @@ class Experiment(object):
       sys.exit(-1)
     # Store new prototypes in model.
     self.prototype_source = 'imprinted'
-    if self.debug:
-      # Convert input source index to corresponding image path.
-      locations = [ (image_files[l[0]],) + l[1:] for l in locations ]
-      self.debug_prototype_locations = locations
-    self.model.s2_kernels = prototypes
+    # Convert input source index to corresponding image path.
+    locations = [ [ (image_files[loc[0]],) + loc[1:]
+        for loc in locs_for_ksize ] for locs_for_ksize in locations ]
+    self.prototype_imprint_locations = locations
+    model.s2_kernels = prototypes
     self.prototype_construction_time = time.time() - start_time
 
-  def MakeUniformRandomS2Prototypes(self, num_prototypes):
+  def MakeUniformRandomS2Prototypes(self, num_prototypes, low = None,
+      high = None):
     """Create a set of S2 prototypes with uniformly random entries.
 
     :param int num_prototypes: The number of S2 prototype arrays to create.
 
     """
+    if low == None:
+      low = 0
+    if high == None:
+      high = 1
     start_time = time.time()
-    shape = (num_prototypes,) + tuple(self.model.s2_kernel_shape)
-    prototypes = np.random.uniform(0, 1, shape)
-    for proto in prototypes:
-      proto /= np.linalg.norm(proto)
-    self.model.s2_kernels = prototypes
+    model = self.model
+    prototypes = []
+    for kshape in model.s2_kernel_shapes:
+      shape = (num_prototypes,) + tuple(kshape)
+      prototypes_for_size = np.random.uniform(low, high, shape)
+      if model.s2_kernels_are_normed:
+        for proto in prototypes_for_size:
+          proto /= np.linalg.norm(proto)
+      prototypes.append(prototypes_for_size)
+    model.s2_kernels = prototypes
     self.prototype_construction_time = time.time() - start_time
     self.prototype_source = 'uniform'
 
@@ -278,8 +321,9 @@ class Experiment(object):
     start_time = time.time()
     if self.model.s2_kernels == None:
       self.ImprintS2Prototypes(num_prototypes)
-    for kernel in self.model.s2_kernels:
-      np.random.shuffle(kernel.flat)
+    for kernels_for_size in self.model.s2_kernels:
+      for kernel in kernels_for_size:
+        np.random.shuffle(kernel.flat)
     self.prototype_construction_time = time.time() - start_time
     self.prototype_source = 'shuffle'
 
@@ -294,15 +338,29 @@ class Experiment(object):
 
     """
     start_time = time.time()
-    # Get ~100k C1 samples, which is approximately 255 prototypes.
-    self.ImprintS2Prototypes(num_prototypes = 255)
-    hist = HistogramSampler(self.model.s2_kernels.flat)
-    size = (num_prototypes,) + self.model.s2_kernel_shape
-    prototypes = hist.Sample(size).astype(
-        util.ACTIVATION_DTYPE)
-    for proto in prototypes:
-      proto /= np.linalg.norm(proto)
-    self.model.s2_kernels = prototypes
+    model = self.model
+    # We treat each kernel size independently. We want the histogram for each
+    # kernel size to be based on ~100k C1 samples minimum. Here, we calculate
+    # the number of imprinted prototypes required to get this.
+    c1_samples_per_prototype = min([ reduce(operator.mul, shape)
+        for shape in model.s2_kernel_shapes ])
+    num_desired_c1_samples = 100000
+    num_imprinted_prototypes = int(num_desired_c1_samples /
+        float(c1_samples_per_prototype))
+    self.ImprintS2Prototypes(num_prototypes = num_imprinted_prototypes)
+    # For each kernel size, build a histogram and sample new prototypes from it.
+    prototypes = []
+    for idx in range(len(model.s2_kernel_shapes)):
+      kernels = model.s2_kernels[idx]
+      shape = model.s2_kernel_shapes[idx]
+      hist = HistogramSampler(kernels.flat)
+      size = (num_prototypes,) + shape
+      prototypes_for_size = hist.Sample(size, dtype = util.ACTIVATION_DTYPE)
+      if model.s2_kernels_are_normed:
+        for proto in prototypes_for_size:
+          proto /= np.linalg.norm(proto)
+      prototypes.append(prototypes_for_size)
+    model.s2_kernels = prototypes
     self.prototype_construction_time = time.time() - start_time
     self.prototype_source = 'histogram'
 
@@ -317,15 +375,31 @@ class Experiment(object):
 
     """
     start_time = time.time()
-    # Get ~100k C1 samples, which is approximately 255 prototypes.
-    self.ImprintS2Prototypes(num_prototypes = 255)
-    mean, std = self.model.s2_kernels.mean(), self.model.s2_kernels.std()
-    size = (num_prototypes,) + self.model.s2_kernel_shape
-    prototypes = np.random.normal(mean, std, size = size).astype(
-        util.ACTIVATION_DTYPE)
-    for proto in prototypes:
-      proto /= np.linalg.norm(proto)
-    self.model.s2_kernels = prototypes
+    model = self.model
+    # We treat each kernel size independently. We want the histogram for each
+    # kernel size to be based on ~100k C1 samples minimum. Here, we calculate
+    # the number of imprinted prototypes required to get this.
+    c1_samples_per_prototype = min([ reduce(operator.mul, shape)
+        for shape in model.s2_kernel_shapes ])
+    num_desired_c1_samples = 100000
+    num_imprinted_prototypes = int(num_desired_c1_samples /
+        float(c1_samples_per_prototype))
+    self.ImprintS2Prototypes(num_prototypes = num_imprinted_prototypes)
+    # For each kernel size, estimate parameters of a normal distribution and
+    # sample from it.
+    prototypes = []
+    for idx in range(len(model.s2_kernel_shapes)):
+      kernels = model.s2_kernels[idx]
+      shape = model.s2_kernel_shapes[idx]
+      mean, std = kernels.mean(), kernels.std()
+      size = (num_prototypes,) + shape
+      prototypes_for_size = np.random.normal(mean, std, size = size).astype(
+          util.ACTIVATION_DTYPE)
+      if model.s2_kernels_are_normed:
+        for proto in prototypes_for_size:
+          proto /= np.linalg.norm(proto)
+      prototypes.append(prototypes_for_size)
+    model.s2_kernels = prototypes
     self.prototype_construction_time = time.time() - start_time
     self.prototype_source = 'normal'
 
@@ -338,26 +412,61 @@ class Experiment(object):
     self.prototype_source = 'manual'
     self.model.s2_kernels = prototypes
 
-  def ComputeFeaturesFromInputStates(self, input_states, block = True):
+  def GetImageFeatures(self, images, resize = None, raw = False,
+      save_all = False, block = True):
     """Return the activity of the model's output layer for a set of images.
 
-    :type input_states: iterable of State
-    :param input_states: Model states containing image data.
-    :rtype: iterable
+    :param input_states: Image paths.
+    :type input_states: str, or iterable of str
+    :param int resize: Fixed length of shortest side. See :class:`InputSource
+       <glimpse.models.misc.InputSource>`.
+    :param bool raw: Whether to return per-image results as a single feature
+       vector or the raw state object.
+    :param bool save_all: Whether resulting states should contain values for
+       all computed layers in the network. See :meth:`BuildLayer
+       <glimpse.models.misc.BaseModel.BuildLayer>`.
+    :param bool block: (experimental) Block while waiting for results.
+    :rtype: iterable of 1D ndarray of float
     :returns: A feature vector for each image.
 
     """
-    L = self.model.LayerClass
-    if self.layer in (L.S2, L.C2, L.IT) and self.model.s2_kernels == None:
-      sys.exit("Please set the S2 prototypes before computing feature vectors "
-          "for layer %s." % self.layer.name)
-    builder = self.model.BuildLayerCallback(self.layer, save_all = False)
+    states = [ self.model.MakeStateFromFilename(fn, resize = resize)
+        for fn in images ]
+    return self.GetStateFeatures(states, raw = raw, save_all = save_all,
+        block = block)
+
+  def GetStateFeatures(self, input_states, raw = False, save_all = False,
+      block = True):
+    """Return the activity of the model's output layer for a set of states.
+
+    :param input_states: Model states containing image data.
+    :type input_states: iterable of State
+    :param bool raw: Whether to return per-image results as a single feature
+       vector or the raw state object.
+    :param bool save_all: Whether resulting states should contain values for
+       all computed layers in the network. See :meth:`BuildLayer
+       <glimpse.models.misc.BaseModel.BuildLayer>`.
+    :param bool block: (experimental) Block while waiting for results.
+    :returns: Features for each image.
+    :rtype: iterable of 1D ndarray of float, or iterable of state (if raw is
+       True)
+
+    """
+    if self.model.s2_kernels == None:
+      lyr = self.model.LayerClass
+      if lyr.IsSublayer(lyr.S2, self.layer):
+        sys.exit("Please set the S2 prototypes before computing feature vectors"
+            " for layer %s." % self.layer.name)
+    builder = self.model.BuildLayerCallback(self.layer, save_all = save_all)
     # Compute model states containing desired features.
     output_states = self.pool.imap(builder, input_states)
     # Look up the activity values for the output layer, and convert them all to
     # a single vector.
-    features = ( util.ArrayListToVector(state[self.layer.id])
-        for state in output_states )
+    layer = self.layer.ident
+    if raw:
+      features = output_states
+    else:
+      features = ( util.FlattenArrays(st[layer]) for st in output_states )
     if block:
       try:
         features = list(features)  # wait for results
@@ -513,11 +622,9 @@ class Experiment(object):
     train_images = util.UngroupLists(self.train_images)
     test_images = util.UngroupLists(self.test_images)
     images = train_images + test_images
-    # Compute features for all images.
-    input_states = [ self.model.MakeStateFromFilename(f, resize = self.resize)
-        for f in images ]
     start_time = time.time()
-    features = self.ComputeFeaturesFromInputStates(input_states, block = True)
+    # Compute features for all images.
+    features = self.GetImageFeatures(images, resize = self.resize, block = True)
     self.compute_feature_time = time.time() - start_time
     # Split results by training/testing set
     train_features, test_features = util.SplitList(features,
@@ -541,11 +648,18 @@ class Experiment(object):
     """
     if self.train_features == None:
       self.ComputeFeatures()
-    svm_model = svm.ScaledSvm(self.scaler)
-    svm_model.Train(self.train_features)
-    self.scaler = svm_model.scaler  # scaler has been trained. save it.
-    self.classifier = svm_model.classifier
-    self.train_results = svm_model.Test(self.train_features)
+    # Prepare the data
+    train_features, train_labels = svm.PrepareFeatures(self.train_features)
+    # Create the SVM classifier with feature scaling.
+    self.classifier = svm.Pipeline([ ('scaler', sklearn.preprocessing.Scaler()),
+        ('svm', sklearn.svm.LinearSVC())])
+    self.classifier.fit(train_features, train_labels)
+    # Evaluate the classifier
+    decision_values = self.classifier.decision_function(train_features)
+    predicted_labels = self.classifier.predict(train_features)
+    accuracy = sklearn.metrics.zero_one_score(train_labels, predicted_labels)
+    self.train_results = dict(decision_values = decision_values,
+        predicted_labels = predicted_labels, accuracy = accuracy)
     return self.train_results['accuracy']
 
   def TestSvm(self):
@@ -557,9 +671,14 @@ class Experiment(object):
     :rtype: float
 
     """
-    svm_model = svm.ScaledSvm(classifier = self.classifier,
-        scaler = self.scaler)
-    self.test_results = svm_model.Test(self.test_features)
+    # Prepare the data
+    test_features, test_labels = svm.PrepareFeatures(self.test_features)
+    # Evaluate the classifier
+    decision_values = self.classifier.decision_function(test_features)
+    predicted_labels = self.classifier.predict(test_features)
+    accuracy = sklearn.metrics.zero_one_score(test_labels, predicted_labels)
+    self.test_results = dict(decision_values = decision_values,
+        predicted_labels = predicted_labels, accuracy = accuracy)
     return self.test_results['accuracy']
 
   def CrossValidateSvm(self):
@@ -571,8 +690,15 @@ class Experiment(object):
     :rtype: float
 
     """
-    test_accuracy = svm.SvmCrossValidate(self.GetFeatures(),
-        num_repetitions = 10, num_splits = 10, scaler = self.scaler)
+    if self.train_features == None:
+      self.ComputeFeatures()
+    features, labels = svm.PrepareFeatures(self.GetFeatures())
+    # Create the SVM classifier with feature scaling.
+    classifier = svm.Pipeline([ ('scaler', sklearn.preprocessing.Scaler()),
+        ('svm', sklearn.svm.LinearSVC())])
+    scores = sklearn.cross_validation.cross_val_score(classifier, features,
+        labels, cv = 10, n_jobs = -1)
+    test_accuracy = scores.mean()
     self.train_results = None
     self.test_results = dict(accuracy = test_accuracy)
     self.cross_validated = True
@@ -606,25 +732,9 @@ class Experiment(object):
 
   def Store(self, root_path):
     """Save the experiment to disk."""
-    # We modify the value of the "classifier" attribute, so cache it.
-    classifier = self.classifier
     pool = self.pool
     self.pool = None  # can't serialize some pools
-    # Use "classifier" attribute to indicate whether LIBSVM classifier is
-    # present.
-    if classifier != None:
-      self.classifier = True
     util.Store(self, root_path)
-    if classifier != None:
-      # Use delayed import of LIBSVM library, so non-SVM methods are always
-      # available.
-      import svmutil
-      svmutil.svm_save_model(root_path + '.svm', classifier)
-    if self.debug and hasattr(pool, 'cluster_stats'):
-      # This is a hackish way to record basic cluster usage information.
-      util.Store(pool.cluster_stats, root_path + '.cluster-stats')
-    # Restore the value of the "classifier" and "pool" attributes.
-    self.classifier = classifier
     self.pool = pool
 
   @staticmethod
@@ -636,18 +746,6 @@ class Experiment(object):
 
     """
     experiment = util.Load(root_path)
-    if experiment.classifier != None:
-      if not isinstance(root_path, basestring):
-        logging.warn("Failed to load SVM model for experiment.")
-      else:
-        model_path = root_path + '.svm'
-        if not os.path.exists(model_path):
-          logging.warn("SVM model not found")
-        else:
-          # Use delayed import of LIBSVM library, so non-SVM methods are always
-          # available.
-          import svmutil
-          experiment.classifier = svmutil.svm_load_model(model_path)
     return experiment
 
 __POOL = None
@@ -696,13 +794,15 @@ def UseCluster(config_file = None, chunksize = None):
 def SetModelClass(model_class = None):
   """Set the model type.
 
-  :param model_class: New model class to use (e.g.,
-     :class:`viz2.Model <glimpse.models.viz2.model.Model>`).
+  :param model_class: Model class to use for future experiments. If a name is
+     given, it is passed to :func:`GetModelClass
+     <glimpse.models.GetModelClass>`.
+  :type model_class: class or str
 
   """
   global __MODEL_CLASS
-  if model_class == None:
-    model_class = viz2.Model  # the default GLIMPSE model
+  if not isinstance(model_class, type):
+    model_class = glimpse.models.GetModelClass(model_class)
   logging.info("Using model type: %s", model_class.__name__)
   __MODEL_CLASS = model_class
   return __MODEL_CLASS
@@ -720,7 +820,7 @@ def SetParams(params = None):
   """
   global __PARAMS
   if params == None:
-    params = GetModelClass().ParamsClass()
+    params = GetModelClass().ParamClass()
   __PARAMS = params
   return __PARAMS
 
@@ -737,10 +837,13 @@ def SetLayer(layer = None):
   """Set the layer from which features will be extracted for the next
   experiment.
 
+  :param layer: Layer from which to compute feature vectors.
+  :type layer: str or :class:`glimpse.models.misc.LayerSpec`
+
   """
   global __LAYER
   if layer == None:
-    layer = GetModelClass().LayerClass.IT
+    layer = GetModelClass().LayerClass.TopLayer()
   elif isinstance(layer, str):
     layer = GetModelClass().LayerClass.FromName(layer)
   __LAYER = layer
@@ -767,12 +870,16 @@ def MakeModel(params = None):
   return __MODEL_CLASS(backends.MakeBackend(), params)
 
 def GetExperiment():
-  """Get the current experiment object."""
+  """Get the current experiment object.
+
+  :rtype: :class:`Experiment`
+
+  """
   if __EXP == None:
     SetExperiment()
   return __EXP
 
-def SetExperiment(model = None, layer = None, scaler = None):
+def SetExperiment(model = None, layer = None):
   """Manually create a new experiment.
 
   This function generally is not called directly. Instead, an experiment object
@@ -781,8 +888,6 @@ def SetExperiment(model = None, layer = None, scaler = None):
   :param model: The Glimpse model to use for processing images.
   :param layer: The layer activity to use for features vectors.
   :type layer: :class:`LayerSpec <glimpse.models.misc.LayerSpec>` or str
-  :param scaler: The feature scaling algorithm, such as the
-     :class:`RangeFeatureScaler <glimpse.util.svm.RangeFeatureScaler>`.
   :returns: The new experiment object.
   :rtype: :class:`Experiment`
 
@@ -797,9 +902,7 @@ def SetExperiment(model = None, layer = None, scaler = None):
     layer = GetLayer()
   elif isinstance(layer, str):
     layer = model.LayerClass.FromName(layer)
-  if scaler == None:
-    scaler = svm.SpheringFeatureScaler()
-  __EXP = Experiment(model, layer, pool = GetPool(), scaler = scaler)
+  __EXP = Experiment(model, layer, pool = GetPool())
   return __EXP
 
 @docstring.copy_dedent(Experiment.ImprintS2Prototypes)
@@ -813,11 +916,12 @@ def ImprintS2Prototypes(num_prototypes):
   return result
 
 @docstring.copy_dedent(Experiment.MakeUniformRandomS2Prototypes)
-def MakeUniformRandomS2Prototypes(num_prototypes):
+def MakeUniformRandomS2Prototypes(num_prototypes, low = None, high = None):
   """" """
   if __VERBOSE:
     print "Making %d uniform random prototypes" % num_prototypes
-  result = GetExperiment().MakeUniformRandomS2Prototypes(num_prototypes)
+  result = GetExperiment().MakeUniformRandomS2Prototypes(num_prototypes, low,
+      high)
   if __VERBOSE:
     print "  done: %s s" % GetExperiment().prototype_construction_time
   return result
@@ -855,8 +959,9 @@ def MakeNormalRandomS2Prototypes(num_prototypes):
 def SetS2Prototypes(prototypes):
   """Set the S2 prototypes from an array or a file.
 
-  :param prototypes: : ndarray of float, str
-     The set of prototypes or a path to a file containing the prototypes.
+  :param prototypes: The set of prototypes or a path to a file containing the
+     prototypes.
+  :type prototypes: str or ndarray of float
 
   """
   if isinstance(prototypes, basestring):
@@ -893,6 +998,21 @@ def ComputeFeatures():
   """" """
   GetExperiment().ComputeFeatures()
 
+@docstring.copy_dedent(Experiment.GetImageFeatures)
+def GetImageFeatures(images, resize = None, raw = False, save_all = False,
+    block = True):
+  """ """
+  if isinstance(images, basestring):
+    images = [ images ]  # Single image was passed, wrap it in a list.
+    single_input = True
+  else:
+    single_input = False
+  results = GetExperiment().GetImageFeatures(images, resize = resize, raw = raw,
+      save_all = save_all, block = block)
+  if single_input:
+    results = results[0]
+  return results
+
 @docstring.copy_dedent(Experiment.CrossValidateSvm)
 def CrossValidateSvm():
   """" """
@@ -951,19 +1071,50 @@ def Verbose(flag = True):
   global __VERBOSE
   __VERBOSE = flag
 
+def GetExampleCorpus():
+  """Get an example image corpus.
+
+  :returns: Corpus path.
+  :rtype: str
+
+  """
+  return os.path.join(os.path.dirname(__file__), '..', 'rc', 'example-corpus')
+
+def GetLargeExampleCorpus():
+  """Get a larger example image corpus.
+
+  :returns: Corpus path.
+  :rtype: str
+
+  """
+  return os.path.join(os.path.dirname(__file__), '..', 'rc',
+      'example-large-corpus')
+
+def GetExampleImage():
+  """Get a single example image.
+
+  :returns: Image path.
+  :rtype: str
+
+  """
+  return os.path.join(GetExampleCorpus(), 'cats', 'Marcus_bed.jpg')
+
+def GetExampleImages():
+  """Get multiple example images.
+
+  :returns: Image paths.
+  :rtype: list of str
+
+  """
+  corpus = GetExampleCorpus()
+  return [ os.path.join(corpus, 'cats', 'Marcus_bed.jpg'), os.path.join(corpus,
+      'dogs', '41-27Monate1.JPG') ]
+
 #### CLI Interface ####
 
-def _GetCliModel(model_name):
-  models = __import__("glimpse.models.%s" % model_name, globals(), locals(),
-      ['Model'], 0)
-  try:
-    return getattr(models, 'Model')
-  except AttributeError:
-    raise util.UsageException("Unknown model (-m): %s" % model_name)
-
 def _InitCli(pool_type = None, cluster_config = None, model_name = None,
-    params = None, edit_params = False, layer = None, debug = False,
-    verbose = 0, resize = None, **opts):
+    params = None, edit_params = False, layer = None, verbose = 0,
+    resize = None, **opts):
   if verbose > 0:
     Verbose(True)
     if verbose > 1:
@@ -981,15 +1132,16 @@ def _InitCli(pool_type = None, cluster_config = None, model_name = None,
       SetPool(pool)
     else:
       raise util.UsageException("Unknown pool type: %s" % pool_type)
-  if model_name != None:
-    SetModelClass(_GetCliModel(model_name))
+  try:
+    SetModelClass(model_name)
+  except ValueError:
+    raise util.UsageException("Unknown model (-m): %s" % model_name)
   SetParams(params)
   SetLayer(layer)
   if edit_params:
     GetParams().configure_traits()
   # At this point, all parameters needed to create Experiment object are set.
   GetExperiment().resize = resize
-  GetExperiment().debug = debug
 
 def _FormatCliResults(svm_decision_values = False, svm_predicted_labels = False,
     **opts):
@@ -1062,7 +1214,6 @@ def CommandLineInterface(**opts):
   _RunCli(**opts)
 
 def main():
-  default_model = "viz2"
   try:
     opts = dict()
     opts['verbose'] = 0
@@ -1090,9 +1241,6 @@ def main():
       elif opt in ('-l', '--layer'):
         opts['layer'] = arg
       elif opt in ('-m', '--model'):
-        # Set the model class
-        if arg == 'default':
-          arg = default_model
         opts['model_name'] = arg
       elif opt in ('-n', '--num-prototypes'):
         opts['num_prototypes'] = int(arg)
